/* 
*   Function:  open files and read them simulteneously or
*              alternatively.  However, altervative reading will reduce
*              the efficiency a lot. We use a buffer to store
*              characters read from a file, so as to avoid making a
*              lot of harddisk access.
*
*        FILE *initFile(fileName): open a file for reading. It returns
*           a FILE pointer
*
*        closeFile(fp): close a file <fp>
*
*        pickupWord(fp,s): put a word into <s> from stream <fp>. If the
*           return value is non-zero, correct.  Otherwise it reaches
*           the end of the stream <fp>. The words are separated by
*           [space], [tab], [return]. Uppercase characters are
*           translated into lowercase characters.

*	 putBufferBack(preFp)
*	 refreshBuffer(fp)
*	 saveBuffer(fp)
*	 recoverBuffer(fp)
*   Masters:   pickupWord -- pick up a word from a stream 
*    	       initFile -- open a file
*    	       closeFile -- close a file
*              SYNOPSIS
*                   #include <stdio.h>
*              
*                   FILE  *initFile( fileName )
*                   char *fileName;
*
*                   int  pickupWord(fp,s)
*                   FILE   *fp;
*                   char  **s;
*
*                   closeFile(fp)
*                   FILE  *fp;
*
* Copyright (C) 1994 Regents of the University of California
* All rights reserved.
*/

#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "netlist.h"

#ifndef   WB_SIZE
#define   WB_SIZE		5000	/*buffer size for keep charcters */
#endif
#ifndef   MAX_WORDLEN
#define   MAX_WORDLEN	50	/* maximum length of a word */
#endif

FILE  *preFp= NULL;		/* keep previous <fp> pointer */
long   preFtell=0;	/* keep previous position of ftell */
long   saveFtell=0;	/* save ftell */

char  wordBuffer[WB_SIZE], saveWordBuffer[WB_SIZE];
char  st[MAX_WORDLEN+1], saveSt[MAX_WORDLEN+1];
int   bufferPtr=0, saveBufferPtr;
int   lineNumber, saveLineNumber;


/*
It opens a readable or writeable file and returns the stream pointer.
*/
FILE  *initFile( filename, rw )
char  *filename,   rw;
{
      FILE   *fopen(),  *fp;

      switch (rw) {
      case 'r': 
            if ( (fp=fopen(filename,"r")) == NULL ) {
	       	printf("ERROR: input file '%s' not exist\n",filename);
                stopAt("initFile", "pickupWord.c:read");
	        return(NULL);
            } else {
 	        lineNumber = 1; /* initialize line number indicator */
                return( fp );
            }
	    break;
      case 'w': 
            if ( (fp=fopen(filename,"w")) == NULL ) {
	       	printf("Error: cannot create output file '%s'\n",filename);
                stopAt("initFile", "pickupWord.c:read");
	        return(NULL);
            } else 
		return( fp );
	    break;
      case 'a': 
            if ( (fp=fopen(filename,"a")) == NULL ) {
	       	printf("Error: cannot append to file '%s'\n",filename);
                stopAt("initFile", "pickupWord.c:read");
	        return(NULL);
            } else 
		return( fp );
	    break;
      default:
	    printf("ERROR: fopen file type %c\n",rw);
	    stopAt("initFile", "pickupWord.c:read");
	    break;
      }
      return(NULL);
}


closeFile(fp)
FILE  *fp;
{
   if (fclose(fp)) {
      printf("Close file ERROR\n");
      errorLine();
      stopAt("closeFile","pickupWord.c:read");
   }
}
  

/* it returns an integer */
int  pickupInt(fp)
FILE   *fp;
{
     char  *st;

     if (pickupWord(fp,&st)) {
     	    return(stringToInt(st));
     } else {
         errorLine();
         stopAt("pickupInt","pickupWord.c:read");
     }
     return(0);
}


/* it returns an integer without refresh the buffer */
int  takeInt(fp)
FILE   *fp;
{
     char  *st;

     if (takeAword(fp,&st)) {
     	    return(stringToInt(st));
     } else {
         errorLine();
         stopAt("takeInt","pickupWord.c:read");
     }
     return(0);
}


/* convert a string to an integer */
int  stringToInt(st)
char   *st;
{
     if (allDigitals(st))
     	 return(atoi(st));
     else {
         errorLine();
	 printf("ERROR: string %s is not an integer\n",st);
         stopAt("stringToInt","pickupWord.c:read");
	 return(-1);
     }
}


/* it returns an integer */
float  pickupFloat(fp)
FILE   *fp;
{
     char  *st;
     double  atof();

     if (pickupWord(fp,&st)) {
	if (allDigitals(st))
     		return((float)atof(st));
	else {
         	errorLine();
         	stopAt("pickupFloat","pickupWord.c:read");
	}
     } else {
         errorLine();
         stopAt("pickupFloat","pickupWord.c:read");
     }
     return((float)0);
}


/* it returns a word each time it is called. PLEASE don't
switch the FILE pointer <fp> so offen. This will lower the
efficiency a lot. It returns 0 if EOF is reached, otherwise
it returns 1.
*/  
int  pickupWord(fp,s)
FILE   *fp;
char  **s;
{
    if (preFp == fp) 
        return(takeAword(fp,s));	/*take a word from buffer wordBuffer[] */
    else {
        if (preFp != NULL) {
	    putBufferBack(preFp);	/* put characters in buffer back to the
				original file */
        }
        preFp = fp;
	refreshBuffer(fp);	/* clean buffer and re_fill buffer */
        return(takeAword(fp,s));	/*take a word from buffer wordBuffer[] */
    }
}


/* bring characters from file <fp> to buffer
*/
refreshBuffer(fp)
FILE *fp;
{
   int  i;

   preFtell = ftell(fp);	/* remember position of file pointer before
				refreshing buffer */
   bufferPtr = 0;
   for (i=0; i<WB_SIZE; i++) {
       if ((wordBuffer[i]=getc(fp)) == EOF)
		 break;
   }
}


/* save buffer and file pointer */
saveBuffer(fp)
FILE *fp;
{
   int  i;

   saveFtell = ftell(fp);	/* save position of file pointer*/
   stcpy(saveSt, st);		/* save <st> */
   saveBufferPtr = bufferPtr; saveLineNumber = lineNumber;
   for (i=0; i<WB_SIZE; i++) {
       saveWordBuffer[i]=wordBuffer[i];
   }
}


/* recover buffer and file pointer */
recoverBuffer(fp)
FILE *fp;
{
   int  i;

   fseek(fp, saveFtell, 0); stcpy(st, saveSt);
   bufferPtr = saveBufferPtr; lineNumber = saveLineNumber;
   for (i=0; i<WB_SIZE; i++) {
       wordBuffer[i] = saveWordBuffer[i];
   }
}


/* put characters in buffer back to the original file <preFp>
*/
putBufferBack(preFp)
FILE   *preFp;
{
     if (fseek(preFp, preFtell+bufferPtr, 0)) {
         printf("seeking improper file ERROR\n");
         errorLine();
	 stopAt("putBufferBack","pickupWord.c:read");
     }
}


/* return 0 for EOF. Otherwise it returns 1 */
int  takeAword( fp, s )
FILE  *fp;
char  **s;
{
    char lowercase();
    int   i=0, returnVal= 1;

    countSkipSpaces(fp); 
    while ( !countSpace(wordBuffer[bufferPtr]) ) {
        if   (wordBuffer[bufferPtr] == EOF){
		returnVal = 0; break;  /* EOF */
	}
			/*take care of special symbol, such as : ; */
        if (specialSymbol(wordBuffer[bufferPtr]))
	      break;
	st[i++] = lowercase(wordBuffer[bufferPtr++]);
        if (bufferPtr >= WB_SIZE)
		refreshBuffer(fp);
        if (i >= MAX_WORDLEN) {
 	     st[20]='\0';
	     printf("\nWord '%s...' is longer than %d ERROR\n",st,MAX_WORDLEN);
	     printf("Please redefine MAX_WORDLEN in file pickupWord.c\n");
             errorLine();
	     stopAt("takeAword","pickupWord.c:read");
        }
    } /* while */
    if (i == 0)
	st[i++] = wordBuffer[bufferPtr++];
    if (bufferPtr >= WB_SIZE)
		refreshBuffer(fp);
    st[i]='\0';  
    *s = st;
    return( returnVal );  /* not EOF */
}


/* it returnss its lowercase if <c> is a uppercase
*/
char lowercase( c )
char c;
{
     /*
     return(((c>0x40)&&(c<0x5b))?c+0x20:c);
     */
     return(c);
}


/* if <c> is a ';', simicolon(c) returns 1. Otherwise returns 0.
*/
int  specialSymbol(c)
char c;
{
    switch ( c ) {
    case '!':
    case '@':
    case '#':
    case '%':
    case '^':
    case '&':
    case '*':
    /* case '/': */
    case ':':
    case ';':  {return(1);  break;}
    default:   {return(0); break;}
    }
    return 0;
}


/* countSkipSpace() which counts lineNumber */
countSkipSpaces(fp)
FILE  *fp;
{
    while (countSpace(wordBuffer[bufferPtr])) { 
	bufferPtr++;
        if (bufferPtr >= WB_SIZE)
		refreshBuffer(fp);
    } /* while */
}


/* countSpace() which counts lineNumber */
int  countSpace(c)
char  c;
{
   if ( (c ==' ')||(c =='	') )
	return( 1 );
   else if (c =='\n') {
	lineNumber++;   return( 1 );
   } else
        return( 0 );
}


errorLine()
{
    printf("Error at line %d\n", lineNumber);
}


/* Stop the program and give out where to stop
<procName> is the name of stopped procedure
<fileName> is the file name of stopped procedure in
*/
stopAt(procName, fileName)
char *procName, *fileName;
{
    printf("Stop at %s() in %s\n",procName,fileName);
    stop();
}


int allDigitals(st)
char   *st;
{
    int  i, dots=0, dashes=0;

    for (i=0; i<strlen(st); i++) {
	 if ((st[i]<'0')||(st[i]>'9'))
	     switch (st[i]) {
	     case '.': dots++; break;
	     case '-': dashes++; break;
	     default:  return 0; break;
			/* non-digital found */
	     }
    }
    if ((dots == 1)&&(dashes == 0))
	    return 4;	/* positive float number */
    else if ((dots == 1)&&(dashes == 1))
	    return 3;	/* negative float number */
    else if ((dots == 0)&&(dashes == 0))
	    return 2;	/* positive integer */
    else if ((dots == 0)&&(dashes == 1))
            return 1;	/* nagative integer */
    else
            return 0;	/* not a number */
}


/* skip <n> words */
skipNwords(fpin, n)
FILE   *fpin;
int  n;
{
   char  *st;
   int  i;

   for (i=0; i<n;  i++) {
   	pickupWord(fpin,&st);
/*
printf(" skip %s\n", st);
*/
   }
}


/* 
It copies the string s2 to s1.
*/
stcpy(s1,s2)
char *s1,*s2;
{ 
	register int i;
	for (i=0; i<=strlen(s2); i++)
		s1[i]=s2[i];
}


/*
Stringcmp(s1,s2) returns an integer. If the returned number is 0,
character string s1 and s2 are same. If the number is -1, s1 < s2.
If the number is 1,  s1>s2.
*/
int stringcmp(char *s1, char *s2)
{
	int  i;

	i = strcmp(s1,s2);
	if (i > 0)
		return(1);
	else if (i == 0)
		return(0);
	else 
	    return(-1);
}


/*
It removes characters after <c>
*/
char   *tailor(st,c)
char *st, c;
{
	int i ;

	for (i=strlen(st); i>=0; i--)
		if (st[i]==c) {
			st[i]='\0'; 
			return(st);
		}

	return(st);
}


stringError(st1,st2)
char  *st1, *st2;
{
	 errorLine();
	 printf("Wrong key word: '%s'!  Put '%s' here\n",st1,st2); 
	 stop();
}


stop()
{
   exit( -1 );
}

