#include "netlist.h"
#include "macro.h"
#include <math.h>
#include <string.h>

#define ALLOC(type, size) (type*) malloc( (unsigned)( sizeof(type) * (size)) )

double * x, * y;

double * px, * py;
int    * orient;
int    * shape;
double * width, * height;

double LLx, LLy, URx, URy;

double a0, w0;
double A, W, cost;

double alpha;

char   student_name[80];
char   circuit_name[80];
char   read_file_name[80];

void GetCircuitName(char * name)
{
    char * ptr;
    
    if (ptr = strrchr(name, '/')) {
        ptr++;
        strcpy(circuit_name, ptr);
    }
    else
        strcpy(circuit_name, name);
}

/* usage:
 *
 *    verify circuit_name alpha student_name 
 */
main(int argc, char *argv[])
{
    int  intg;
    
    if (argc == 4) {
        
        read_netlist(argv[1]);		/* read in netlist */
        
        strcpy(student_name, argv[3]);
        
        sscanf(argv[2], "%d", & intg);
        if (intg == 25) alpha = 0.25;
        else alpha = 0.75;
        
        GetSpace();
        
        GetCircuitName(argv[1]);
        
        sprintf(read_file_name, "%s.%s.%-d.out", student_name, circuit_name, intg);
        
        ReadSolution(read_file_name);
        
        FindParameters();
        
        EvaluateSolution();
        
        cost = alpha * A / a0 + (1.0 - alpha) * W / w0;
        
        fprintf(stderr, "%10s %12s %12s %12s\n", "name", "area", "wirelen", "cost");
        printf("%10s %12.2e %12.2e %12.2lf\n", student_name, A, W, cost);
    }
    else
        fprintf(stderr, "Sorry, no arguments given\n");
    return 0;
}

Cell * FindCellPad(char * name, int index)
{
    int i;
    
    for(i=0; i<netlist->nc; i++) {
        if (netlist->cells[i].index == index && 
            strcmp(name, netlist->cells[i].name) == 0)
            return & netlist->cells[i];
    }
    
    for(i=0; i<netlist->np; i++) {
        if (netlist-> pads[i].index == index && 
            strcmp(name, netlist-> pads[i].name) == 0)
            return & netlist-> pads[i];
    }

    /* we shouldn't go here */
    fprintf(stderr, "Sorry, cannot find cell (%s).\n", name);
    exit(-1);
}

ReadSolution(char * file_name)
{
    FILE * fp;
    int i, index, ort, shp;
    char cpname[80];
    double myx, myy;
    Cell * cellpad;
    
    fp = fopen(file_name, "r");
    if (! fp) {
        fprintf(stderr, "Cannot open file (%s) for reading...\n", file_name);
        exit(-1);
    }
    
    for(i=0; i<netlist->nc + netlist->np; i++) {
		if ( fscanf(fp, "%s%d%lf%lf%d%d", cpname, & index, 
               & myx, & myy, & ort, & shp) != 6) {
			   fprintf(stderr, "Input file error: something missing...\n");
			   break;
		}

        cellpad = FindCellPad(cpname, index);
        if (isAcell (cellpad) ) {
            x[ cellpad->index ] = myx;
            y[ cellpad->index ] = myy;
            orient[ cellpad->index ] = ort;
            shape[ cellpad->index ] = shp;
        }
        else {
            px[ cellpad->index ] = myx;
            py[ cellpad->index ] = myy;
        }
    }
    
    fclose(fp);
}

GetSpace()
{
    x = ALLOC(double, netlist->nc);
    y = ALLOC(double, netlist->nc);
    px = ALLOC(double, netlist->np);
    py = ALLOC(double, netlist->np);
    
    orient = ALLOC(int, netlist->nc);
    shape = ALLOC(int, netlist->nc);
    width = ALLOC(double, netlist->nc);
    height = ALLOC(double, netlist->nc);
}

FindParameters()
{
    int i, j;
    double min, area, w, h, sum = 0.0;
    Cell * cell;
    
    for(i=0; i<netlist->nc; i++) {
        cell = & netlist->cells[i];
        
        min = 1.0e30;
        for(j=0; j<cell->ns; j++) {
			w = cell->shapes[j].urx - cell->shapes[j].llx;
			h = cell->shapes[j].ury - cell->shapes[j].lly;
            area = w * h;
            if (area < min) min = area;
        }
        
        sum += min;
    }
    
    a0 = sum;
    
    w0 = 5.0 * sqrt(a0) * sqrt(1.0 * netlist->nn);
}

void FindBoundary(int i, double * llx, double * lly, 
                  double * urx, double * ury)
{
    double cx, cy, w, h;
    
    cx = x[i];
    cy = y[i];
    
    if (orient[i] < 4) {
        w = netlist->cells[i].shapes[ shape[i] ]. urx -
            netlist->cells[i].shapes[ shape[i] ]. llx;
        h = netlist->cells[i].shapes[ shape[i] ]. ury -
            netlist->cells[i].shapes[ shape[i] ]. lly;
    }
    else {
        h = netlist->cells[i].shapes[ shape[i] ]. urx -
            netlist->cells[i].shapes[ shape[i] ]. llx;
        w = netlist->cells[i].shapes[ shape[i] ]. ury -
            netlist->cells[i].shapes[ shape[i] ]. lly;
    }
    
    * llx = cx - w / 2.0;
    * urx = cx + w / 2.0;
    * lly = cy - h / 2.0;
    * ury = cy + h / 2.0;
}

void ComputeArea()
{
    int i;
    double wl = 0.0;
    double llx, lly, urx, ury;
    double clx, cly, cux, cuy;
    
    /* first, find the bounding box */
    for(i=0; i<netlist->nc; i++) {
        FindBoundary(i, &clx, &cly, &cux, &cuy);
        if (i == 0) {
            llx = clx, lly = cly;
            urx = cux, ury = cuy;
        }
        else {
            if (clx < llx) llx = clx;
            if (cly < lly) lly = cly;
            if (cux > urx) urx = cux;
            if (cuy > ury) ury = cuy;
        }
    }    

    llx -= 4, lly -= 4;
    urx += 4, ury += 4;

	LLx = llx;
	URx = urx;
	LLy = lly;
	URy = ury;

    A = (urx - llx) * (ury - lly);
}

double ComputeWireLength( Net * net )
{
    double wl = 0.0;
    int    i;
    double llx, lly, urx, ury, myx, myy, offx, offy;
    Pin * pin;
	int   nrCell;
    
	if (net->np < 2) return 0.0;

	urx = ury = llx = lly = 0;	/* to avoid something strange */

	nrCell = 0;
    for(i=0; i<net->np; i++) {
        pin = net->pins[i];
        if (isAcell(pin->cell)) {
            /* don't consider equivalent pins here */
            myx = x[ pin->cell->index ];
            myy = y[ pin->cell->index ];
            
            /* offset: ??? */
            offx = pin->x;
            offy = pin->y;
            switch( orient[i] ) {
            case 0: 
                myx += offx;
                myy += offy;
                break;
            case 1:
                myx += offx;
                myy -= offy;
                break;
            case 2:
                myx -= offx;
                myy += offy;
                break;
            case 3:
                myx -= offx;
                myy -= offy;
            case 4:
                myx -= offy;
                myy -= offx;
                break;
            case 5:
                myx += offy;
                myy += offx;
                break;
            case 6:
                myx -= offy;
                myy += offx;
                break;
                
            case 7:
                myx += offy;
                myy -= offx;
                break;
            }
        	if (nrCell == 0) {
           		llx = urx = myx;
            	lly = ury = myy;
        	}
        	else {
            	if (myx < llx) llx = myx;
            	if (myx > urx) urx = myx;
            	if (myy < lly) lly = myy;
           		if (myy > ury) ury = myy;
        	}

			nrCell ++;
        }
#if 0
            myx = px[ pin->cell->index ];
            myy = py[ pin->cell->index ];
#endif 0
    }
    
	if (nrCell == net->np)
	    return (urx - llx) + (ury - lly);

    for(i=0; i<net->np; i++) {
        pin = net->pins[i];
        if (isAcell(pin->cell)) continue;
		switch( pin->cell->type ) {
		case LEFT_PAD:   llx = LLx; break;
		case RIGHT_PAD:  urx = URx; break;
		case BOTTOM_PAD: lly = LLy; break;
		case TOP_PAD:    ury = URy; break;
		}
	}
    return (urx - llx) + (ury - lly);
}

EvaluateSolution()
{
    int i;
    double cost;
    
    ComputeArea();

    W = 0.0;
    
    for(i=0; i<netlist->nn; i++) {
        W += ComputeWireLength( & netlist->nets[i] );
    }
}


